# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from stestr.tests import base
from stestr import utils


class TestUtils(base.TestCase):
    def test_cleanup_test_name_defaults(self):
        test_id_no_attrs = 'test.TestThing.test_thing'
        test_id_with_attrs = 'test.TestThing.test_thing[attr1,attr2,att3]'
        test_id_with_scenario = 'test.TestThing.test_thing(mysql)'
        test_id_with_attrs_and_scenario = ('test.TestThing.test_thing[attr]'
                                           '(mysql)')
        result_no_attrs = utils.cleanup_test_name(test_id_no_attrs)
        self.assertEqual(test_id_no_attrs, result_no_attrs)
        result_with_attrs = utils.cleanup_test_name(test_id_with_attrs)
        self.assertEqual(test_id_no_attrs, result_with_attrs)
        result_with_scenario = utils.cleanup_test_name(test_id_with_scenario)
        self.assertEqual(test_id_with_scenario, result_with_scenario)
        result_with_attr_and_scenario = utils.cleanup_test_name(
            test_id_with_attrs_and_scenario)
        self.assertEqual(test_id_with_scenario, result_with_attr_and_scenario)

    def test_cleanup_test_name_leave_attrs(self):
        test_id_no_attrs = 'test.TestThing.test_thing'
        test_id_with_attrs = 'test.TestThing.test_thing[attr1,attr2,att3]'
        test_id_with_scenario = 'test.TestThing.test_thing(mysql)'
        test_id_with_attrs_and_scenario = ('test.TestThing.test_thing[attr]'
                                           '(mysql)')
        result_no_attrs = utils.cleanup_test_name(test_id_no_attrs,
                                                  strip_tags=False)
        self.assertEqual(test_id_no_attrs, result_no_attrs)
        result_with_attrs = utils.cleanup_test_name(test_id_with_attrs,
                                                    strip_tags=False)
        self.assertEqual(test_id_with_attrs, result_with_attrs)
        result_with_scenario = utils.cleanup_test_name(test_id_with_scenario,
                                                       strip_tags=False)
        self.assertEqual(test_id_with_scenario, result_with_scenario)
        result_with_attr_and_scenario = utils.cleanup_test_name(
            test_id_with_attrs_and_scenario, strip_tags=False)
        self.assertEqual(test_id_with_attrs_and_scenario,
                         result_with_attr_and_scenario)

    def test_cleanup_test_name_strip_scenario_and_attrs(self):
        test_id_no_attrs = 'test.TestThing.test_thing'
        test_id_with_attrs = 'test.TestThing.test_thing[attr1,attr2,att3]'
        test_id_with_scenario = 'test.TestThing.test_thing(mysql)'
        test_id_with_attrs_and_scenario = ('test.TestThing.test_thing[attr]'
                                           '(mysql)')
        result_no_attrs = utils.cleanup_test_name(test_id_no_attrs,
                                                  strip_scenarios=True)
        self.assertEqual(test_id_no_attrs, result_no_attrs)
        result_with_attrs = utils.cleanup_test_name(test_id_with_attrs,
                                                    strip_scenarios=True)
        self.assertEqual(test_id_no_attrs, result_with_attrs)
        result_with_scenario = utils.cleanup_test_name(test_id_with_scenario,
                                                       strip_scenarios=True)
        self.assertEqual(test_id_no_attrs, result_with_scenario)
        result_with_attr_and_scenario = utils.cleanup_test_name(
            test_id_with_attrs_and_scenario, strip_scenarios=True)
        self.assertEqual(test_id_no_attrs,
                         result_with_attr_and_scenario)

    def test_cleanup_test_name_strip_scenario(self):
        test_id_no_attrs = 'test.TestThing.test_thing'
        test_id_with_attrs = 'test.TestThing.test_thing[attr1,attr2,att3]'
        test_id_with_scenario = 'test.TestThing.test_thing(mysql)'
        test_id_with_attrs_and_scenario = ('test.TestThing.test_thing[attr]'
                                           '(mysql)')
        result_no_attrs = utils.cleanup_test_name(test_id_no_attrs,
                                                  strip_scenarios=True,
                                                  strip_tags=False)
        self.assertEqual(test_id_no_attrs, result_no_attrs)
        result_with_attrs = utils.cleanup_test_name(test_id_with_attrs,
                                                    strip_scenarios=True,
                                                    strip_tags=False)
        self.assertEqual(test_id_with_attrs, result_with_attrs)
        result_with_scenario = utils.cleanup_test_name(test_id_with_scenario,
                                                       strip_scenarios=True,
                                                       strip_tags=False)
        self.assertEqual(test_id_no_attrs, result_with_scenario)
        result_with_attr_and_scenario = utils.cleanup_test_name(
            test_id_with_attrs_and_scenario, strip_scenarios=True,
            strip_tags=False)
        self.assertEqual('test.TestThing.test_thing[attr]',
                         result_with_attr_and_scenario)



# Copyright 2019 Matthew Treinish
# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2007
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from functools import partial
import os
import sys

from subunit import StreamResultToBytes
from subunit.test_results import AutoTimingTestResultDecorator
from testtools import ExtendedToStreamDecorator

from stestr.subunit_runner import program


class SubunitTestRunner(object):
    def __init__(self, failfast=False, tb_locals=False, stdout=sys.stdout):
        """Create a Test Runner.

        :param failfast: Stop running tests at the first failure.
        :param stdout: Output stream parameter, defaults to sys.stdout
        :param tb_locals: If set true local variables will be shown

        Either stream or stdout can be supplied, and stream will take
        precedence.
        """
        self.failfast = failfast
        self.stream = stdout
        self.tb_locals = tb_locals

    def run(self, test):
        "Run the given test case or test suite."
        result, _ = self._list(test)
        result = ExtendedToStreamDecorator(result)
        result = AutoTimingTestResultDecorator(result)
        if self.failfast is not None:
            result.failfast = self.failfast
            result.tb_locals = self.tb_locals
        result.startTestRun()
        try:
            test(result)
        finally:
            result.stopTestRun()
        return result

    def list(self, test, loader=None):
        "List the test."
        result, errors = self._list(test)
        if loader is not None:
            errors = loader.errors
        if errors:
            failed_descr = '\n'.join(errors).encode('utf8')
            result.status(file_name="import errors", runnable=False,
                          file_bytes=failed_descr,
                          mime_type="text/plain;charset=utf8")
            sys.exit(2)

    def _list(self, test):
        test_ids, errors = program.list_test(test)
        try:
            fileno = self.stream.fileno()
        except Exception:
            fileno = None
        if fileno is not None:
            stream = os.fdopen(fileno, 'wb', 0)
        else:
            stream = self.stream
        result = StreamResultToBytes(stream)
        for test_id in test_ids:
            result.status(test_id=test_id, test_status='exists')
        return result, errors


def main():
    runner = SubunitTestRunner
    program.TestProgram(
        module=None, argv=sys.argv,
        testRunner=partial(runner, stdout=sys.stdout))


if __name__ == '__main__':
    main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Interact with the run history in a repository."""

import functools
import sys

from cliff import command
from cliff import lister
import subunit
import testtools

from stestr import output
from stestr.repository import abstract
from stestr.repository import util
from stestr import results
from stestr import subunit_trace
from stestr import user_config


class HistoryList(lister.Lister):
    """List the the run history in a repository."""

    def get_parser(self, prog_name):
        list_parser = super().get_parser(prog_name)
        list_parser.add_argument('--show-metadata', '-m', action='store_true',
                                 help="Show metadata associated with runs in "
                                      "output table")
        return list_parser

    def take_action(self, parsed_args):
        args = parsed_args
        user_conf = user_config.get_user_config(self.app_args.user_config)
        show_metadata = args.show_metadata
        if getattr(user_conf, 'history_list', False):
            if user_conf.history_show.get('show-metadata',
                                          None) is not None:
                show_metadata = args.show_metadata
        return history_list(repo_url=self.app_args.repo_url,
                            show_metadata=show_metadata)


class HistoryShow(command.Command):
    """Show the contents of a single run in the history"""

    def get_parser(self, prog_name):
        show_parser = super().get_parser(prog_name)
        show_parser.add_argument(
            "--subunit", action="store_true",
            default=False, help="Show output as a subunit stream.")
        show_parser.add_argument("--no-subunit-trace", action='store_true',
                                 default=False,
                                 help="Disable output with the subunit-trace "
                                      "output filter")
        show_parser.add_argument('--force-subunit-trace', action='store_true',
                                 default=False,
                                 help='Force subunit-trace output regardless '
                                      'of any other options or config '
                                      'settings')
        show_parser.add_argument('--color', action='store_true', default=False,
                                 help='Enable color output in the '
                                      'subunit-trace output, if subunit-trace '
                                      'output is enabled. (this is the '
                                      'default). If subunit-trace is disabled '
                                      'this does nothing.')
        show_parser.add_argument('--suppress-attachments', action='store_true',
                                 dest='suppress_attachments',
                                 help='If set do not print stdout or stderr '
                                      'attachment contents on a successful '
                                      'test execution')
        show_parser.add_argument('--all-attachments', action='store_true',
                                 dest='all_attachments',
                                 help='If set print all text attachment '
                                      'contents on a successful test '
                                      'execution')
        show_parser.add_argument('--show-binary-attachments',
                                 action='store_true',
                                 dest='show_binary_attachments',
                                 help='If set, show non-text attachments. '
                                      'This is generally only useful for '
                                      'debug purposes.')
        show_parser.add_argument('run_id', nargs='?', default=None,
                                 help='The run id to show the results from '
                                      'if not specified the last run will be '
                                      'shown (which is equivalent to '
                                      "'stestr last')")
        return show_parser

    def take_action(self, parsed_args):
        args = parsed_args
        user_conf = user_config.get_user_config(self.app_args.user_config)
        if args.suppress_attachments and args.all_attachments:
            msg = ("The --suppress-attachments and --all-attachments "
                   "options are mutually exclusive, you can not use both "
                   "at the same time")
            print(msg)
            sys.exit(1)
        if getattr(user_conf, 'history_show', False):
            if not user_conf.history_show.get('no-subunit-trace'):
                if not args.no_subunit_trace:
                    pretty_out = True
                else:
                    pretty_out = False
            else:
                pretty_out = False
            pretty_out = args.force_subunit_trace or pretty_out
            color = args.color or user_conf.history_show.get('color',
                                                             False)
            suppress_attachments_conf = user_conf.history_show.get(
                'suppress-attachments', False)
            all_attachments_conf = user_conf.history_show.get(
                'all-attachments', False)
            if not args.suppress_attachments and not args.all_attachments:
                suppress_attachments = suppress_attachments_conf
                all_attachments = all_attachments_conf
            elif args.suppress_attachments:
                all_attachments = False
                suppress_attachments = args.suppress_attachments
            elif args.all_attachments:
                suppress_attachments = False
                all_attachments = args.all_attachments
        else:
            pretty_out = args.force_subunit_trace or not \
                args.no_subunit_trace
            color = args.color
            suppress_attachments = args.suppress_attachments
            all_attachments = args.all_attachments
        return history_show(
            args.run_id,
            repo_url=self.app_args.repo_url,
            subunit_out=args.subunit,
            pretty_out=pretty_out, color=color,
            suppress_attachments=suppress_attachments,
            all_attachments=all_attachments,
            show_binary_attachments=args.show_binary_attachments)


class HistoryRemove(command.Command):
    """Remove a run from the history"""

    def get_parser(self, prog_name):
        remove_parser = super().get_parser(prog_name)
        remove_parser.add_argument('run_id',
                                   help='The run id to remove from the '
                                        'repository. Also the string "all" '
                                        'can be used to remove all runs in '
                                        'the history')
        return remove_parser

    def take_action(self, parsed_args):
        args = parsed_args
        history_remove(args.run_id,
                       repo_url=self.app_args.repo_url)


start_times = None
stop_times = None


def _get_run_details(stream_file, stdout):
    stream = subunit.ByteStreamToStreamResult(stream_file,
                                              non_subunit_name='stdout')
    global start_times
    global stop_times
    start_times = []
    stop_times = []

    def collect_data(stream, test):
        global start_times
        global stop_times
        start_times.append(test['timestamps'][0])
        stop_times.append(test['timestamps'][1])

    outcomes = testtools.StreamToDict(functools.partial(collect_data, stdout))
    summary = testtools.StreamSummary()
    result = testtools.CopyStreamResult([outcomes, summary])
    result = testtools.StreamResultRouter(result)
    cat = subunit.test_results.CatFiles(stdout)
    result.add_rule(cat, 'test_id', test_id=None)
    result.startTestRun()
    try:
        stream.run(result)
    finally:
        result.stopTestRun()
    successful = results.wasSuccessful(summary)
    if start_times and stop_times:
        start_time = min(start_times)
        stop_time = max(stop_times)
        run_time = subunit_trace.get_duration([start_time, stop_time])
    else:
        run_time = '---'
        successful = '---'
        start_time = '---'
    return {'passed': successful, 'runtime': run_time, 'start': start_time}


def history_list(repo_url=None, show_metadata=False,
                 stdout=sys.stdout):
    """Show a list of runs in a repository

    Note this function depends on the cwd for the repository if `repo_url` is
    not specified it will use the repository located at CWD/.stestr

    :param str repo_url: The url of the repository to use.
    :param bool show_metadata: If set to ``True`` a column with any metadata
        for a run will be included in the output.
    :param file stdout: The output file to write all output to. By default
         this is sys.stdout

    :return return_code: The exit code for the command. 0 for success and > 0
        for failures.
    :rtype: int
    """

    field_names = ()
    if show_metadata:
        field_names = ('Run ID', 'Passed', 'Runtime', 'Date', 'Metadata')
    else:
        field_names = ('Run ID', 'Passed', 'Runtime', 'Date')
    try:
        repo = util.get_repo_open(repo_url=repo_url)
    except abstract.RepositoryNotFound as e:
        stdout.write(str(e) + '\n')
        return 1
    try:
        run_ids = repo.get_run_ids()
    except KeyError as e:
        stdout.write(str(e) + '\n')
        return 1
    rows = []
    for run_id in run_ids:
        run = repo.get_test_run(run_id)
        stream = run.get_subunit_stream()
        data = _get_run_details(stream, stdout)
        if show_metadata:
            rows.append((run_id, data['passed'], data['runtime'],
                         data['start'], run.get_metadata()))
        else:
            rows.append((run_id, data['passed'], data['runtime'],
                         data['start']))

    return (field_names, rows)


def history_show(run_id, repo_url=None, subunit_out=False,
                 pretty_out=True, color=False, stdout=sys.stdout,
                 suppress_attachments=False, all_attachments=False,
                 show_binary_attachments=False):
    """Show a run loaded into a repository

    This function will print the results from the last run in the repository
    to STDOUT. It can optionally print the subunit stream for the last run
    to STDOUT if the ``subunit`` option is set to true.

    Note this function depends on the cwd for the repository if `repo_url` is
    not specified it will use the repository located at CWD/.stestr

    :param str run_id: The run id to show
    :param str repo_url: The url of the repository to use.
    :param bool subunit_out: Show output as a subunit stream.
    :param pretty_out: Use the subunit-trace output filter.
    :param color: Enable colorized output with the subunit-trace output filter.
    :param bool subunit: Show output as a subunit stream.
    :param file stdout: The output file to write all output to. By default
         this is sys.stdout
    :param bool suppress_attachments: When set true attachments subunit_trace
        will not print attachments on successful test execution.
    :param bool all_attachments: When set true subunit_trace will print all
        text attachments on successful test execution.
    :param bool show_binary_attachments: When set to true, subunit_trace will
        print binary attachments in addition to text attachments.

    :return return_code: The exit code for the command. 0 for success and > 0
        for failures.
    :rtype: int
    """
    try:
        repo = util.get_repo_open(repo_url=repo_url)
    except abstract.RepositoryNotFound as e:
        stdout.write(str(e) + '\n')
        return 1
    try:
        if run_id:
            run = repo.get_test_run(run_id)
        else:
            run = repo.get_latest_run()
    except KeyError as e:
        stdout.write(str(e) + '\n')
        return 1

    if subunit_out:
        stream = run.get_subunit_stream()
        output.output_stream(stream, output=stdout)
        # Exits 0 if we successfully wrote the stream.
        return 0
    case = run.get_test()
    try:
        if run_id:
            previous_run = int(run_id) - 1
        else:
            previous_run = repo.get_test_run(repo.latest_id() - 1)
    except KeyError:
        previous_run = None
    failed = False
    if not pretty_out:
        output_result = results.CLITestResult(run.get_id, stdout,
                                              previous_run)
        summary = output_result.get_summary()
        output_result.startTestRun()
        try:
            case.run(output_result)
        finally:
            output_result.stopTestRun()
        failed = not results.wasSuccessful(summary)
    else:
        stream = run.get_subunit_stream()
        failed = subunit_trace.trace(
            stream, stdout, post_fails=True, color=color,
            suppress_attachments=suppress_attachments,
            all_attachments=all_attachments,
            show_binary_attachments=show_binary_attachments)
    if failed:
        return 1
    else:
        return 0


def history_remove(run_id, repo_url=None, stdout=sys.stdout):
    """Remove a run from a repository

    Note this function depends on the cwd for the repository if `repo_url` is
    not specified it will use the repository located at CWD/.stestr

    :param str run_id: The run id to remove from the repository. Also, can be
        set to ``all`` which will remove all runs from the repository.
    :param str repo_url: The url of the repository to use.
    :param file stdout: The output file to write all output to. By default
         this is sys.stdout

    :return return_code: The exit code for the command. 0 for success and > 0
        for failures.
    :rtype: int
    """
    try:
        repo = util.get_repo_open(repo_url=repo_url)
    except abstract.RepositoryNotFound as e:
        stdout.write(str(e) + '\n')
        return 1
    if run_id == 'all':
        try:
            run_ids = repo.get_run_ids()
        except KeyError as e:
            stdout.write(str(e) + '\n')
            return 1
        for run_id in run_ids:
            repo.remove_run_id(run_id)
    else:
        try:
            repo.remove_run_id(run_id)
        except KeyError as e:
            stdout.write(str(e) + '\n')
            return 1
    return 0



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io

from stestr import output
from stestr.tests import base


class TestOutput(base.TestCase):

    def test_output_table(self):
        table = [['Header 1', 'Header 2', 'Header 999'],
                 [1, '0000000002', 'foo'],
                 ['bar', 6, 'This is a content.']]
        expected = \
            "Header 1  Header 2    Header 999\n" \
            "--------  ----------  ------------------\n" \
            "1         0000000002  foo\n" \
            "bar       6           This is a content.\n"
        with io.StringIO() as f:
            output.output_table(table, f)
            actual = f.getvalue()
            self.assertEqual(expected, actual)

    def test_output_tests(self):
        class Test:
            def __init__(self, i):
                self.i = i

            def id(self):
                return self.i

        tests = [Test('a'), Test('b'), Test('foo')]
        expected = "a\nb\nfoo\n"
        with io.StringIO() as f:
            output.output_tests(tests, f)
            actual = f.getvalue()
            self.assertEqual(expected, actual)

    def test_output_summary_passed(self):
        expected = 'Ran 10 (+5) tests in 1.100s (+0.100s)\n' \
            'PASSED (id=99 (+1), id=100 (+2))\n'
        with io.StringIO() as f:
            output.output_summary(
                successful=True, tests=10, tests_delta=5,
                time=1.1, time_delta=0.1,
                values=[('id', 99, 1), ('id', '100', 2)],
                output=f)
            actual = f.getvalue()
            self.assertEqual(expected, actual)

    def test_output_summary_failed(self):
        expected = 'Ran 10 (+5) tests in 1.100s (+0.100s)\n' \
            'FAILED (id=99 (+1), id=100 (+2))\n'
        with io.StringIO() as f:
            output.output_summary(
                successful=False, tests=10, tests_delta=5,
                time=1.1, time_delta=0.1,
                values=[('id', 99, 1), ('id', '100', 2)],
                output=f)
            actual = f.getvalue()
            self.assertEqual(expected, actual)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import re
from unittest import mock

from stestr import selection
from stestr.tests import base


class TestSelection(base.TestCase):
    def test_filter_tests_no_filter(self):
        test_list = ['a', 'b', 'c']
        result = selection.filter_tests(None, test_list)
        self.assertEqual(test_list, result)

    def test_filter_tests(self):
        test_list = ['a', 'b', 'c']
        result = selection.filter_tests(['a'], test_list)
        self.assertEqual(['a'], result)

    def test_filter_invalid_regex(self):
        test_list = ['a', 'b', 'c']
        with mock.patch('sys.exit', side_effect=ImportError) as mock_exit:
            self.assertRaises(ImportError, selection.filter_tests,
                              ['fake_regex_with_bad_part[The-BAD-part]'],
                              test_list)
            mock_exit.assert_called_once_with(5)


class TestExclusionReader(base.TestCase):

    def test_exclusion_reader(self):
        exclude_list = io.StringIO()
        for i in range(4):
            exclude_list.write('fake_regex_%s\n' % i)
            exclude_list.write('fake_regex_with_note_%s # note\n' % i)
        exclude_list.seek(0)
        with mock.patch('builtins.open',
                        return_value=exclude_list):
            result = selection.exclusion_reader('fake_path')
        self.assertEqual(2 * 4, len(result))
        note_cnt = 0
        # not assuming ordering, mainly just testing the type
        for r in result:
            self.assertEqual(r[2], [])
            if r[1] == 'note':
                note_cnt += 1
            self.assertIn('search', dir(r[0]))  # like a compiled regexp
        self.assertEqual(note_cnt, 4)

    def test_invalid_regex(self):
        exclude_list = io.StringIO()
        exclude_list.write("fake_regex_with_bad_part[The-BAD-part]")
        exclude_list.seek(0)
        with mock.patch('builtins.open',
                        return_value=exclude_list):
            with mock.patch('sys.exit') as mock_exit:
                selection.exclusion_reader('fake_path')
                mock_exit.assert_called_once_with(5)


class TestConstructList(base.TestCase):
    def test_simple_re(self):
        test_lists = ['fake_test(scen)[tag,bar])', 'fake_test(scen)[egg,foo])',
                      'fake_test(necs)[tag,bar])', 'fake_test(necs)[egg,foo])',
                      'fake_test(nnnn)[foo,bar])', 'fake_test(nnnn)[foo,foo])']
        result = selection.construct_list(test_lists, regexes=['foo'])
        # Order must be preserved
        expected = test_lists[:]
        del expected[2]
        del expected[0]
        self.assertEqual(expected, result)

    def test_simple_re_randomized(self):
        test_lists = ['fake_test(scen)[tag,bar])', 'fake_test(scen)[egg,foo])',
                      'fake_test(necs)[tag,bar])', 'fake_test(necs)[egg,foo])',
                      'fake_test(nnnn)[foo,bar])', 'fake_test(nnnn)[foo,foo])']
        result = selection.construct_list(test_lists, regexes=['foo'],
                                          randomize=True)
        expected_names = test_lists[:]
        del expected_names[2]
        del expected_names[0]
        # Order is randomized
        self.assertNotEqual(expected_names, result)
        self.assertEqual(set(expected_names), set(result))

    def test_simple_exclusion_re(self):
        test_lists = ['fake_test(scen)[tag,bar])', 'fake_test(scen)[egg,foo])']
        result = selection.construct_list(test_lists, exclude_regex='foo')
        self.assertEqual(list(result), ['fake_test(scen)[tag,bar])'])

    def test_invalid_exclusion_re(self):
        test_lists = ['fake_test(scen)[tag,bar])', 'fake_test(scen)[egg,foo])']
        invalid_regex = "fake_regex_with_bad_part[The-BAD-part]"
        with mock.patch('sys.exit', side_effect=ImportError) as exit_mock:
            self.assertRaises(ImportError, selection.construct_list,
                              test_lists, exclude_regex=invalid_regex)
            exit_mock.assert_called_once_with(5)

    def test_exclusion_list(self):
        exclude_list = [(re.compile('foo'), 'foo not liked', [])]
        test_lists = ['fake_test(scen)[tag,bar])', 'fake_test(scen)[egg,foo])']
        with mock.patch('stestr.selection.exclusion_reader',
                        return_value=exclude_list):
            result = selection.construct_list(test_lists,
                                              exclude_list='file',
                                              regexes=['fake_test'])
        self.assertEqual(list(result), ['fake_test(scen)[tag,bar])'])

    def test_inclusion_list(self):
        include_list = [re.compile('fake_test1'), re.compile('fake_test2')]
        test_lists = ['fake_test1[tg]', 'fake_test2[tg]', 'fake_test3[tg]']
        include_getter = 'stestr.selection._get_regex_from_include_list'
        with mock.patch(include_getter,
                        return_value=include_list):
            result = selection.construct_list(test_lists,
                                              include_list='file')
        self.assertEqual(set(result),
                         {'fake_test1[tg]', 'fake_test2[tg]'})

    def test_inclusion_list_invalid_regex(self):
        include_list = io.StringIO()
        include_list.write("fake_regex_with_bad_part[The-BAD-part]")
        include_list.seek(0)
        with mock.patch('builtins.open',
                        return_value=include_list):
            with mock.patch('sys.exit') as mock_exit:
                selection._get_regex_from_include_list('fake_path')
                mock_exit.assert_called_once_with(5)

    def test_inclusion_exclusion_list_re(self):
        include_list = [re.compile('fake_test1'), re.compile('fake_test2')]
        test_lists = ['fake_test1[tg]', 'fake_test2[spam]',
                      'fake_test3[tg,foo]', 'fake_test4[spam]']
        exclude_list = [(re.compile('spam'), 'spam not liked', [])]
        include_getter = 'stestr.selection._get_regex_from_include_list'
        with mock.patch(include_getter,
                        return_value=include_list):
            with mock.patch('stestr.selection.exclusion_reader',
                            return_value=exclude_list):
                result = selection.construct_list(
                    test_lists, exclude_list='exclude_file',
                    include_list='include_file', regexes=['foo'])
        self.assertEqual(set(result),
                         {'fake_test1[tg]', 'fake_test3[tg,foo]'})

    def test_overlapping_exclude_regex(self):

        exclude_list = [(re.compile('compute.test_keypairs.KeypairsTestV210'),
                         '', []),
                        (re.compile('compute.test_keypairs.KeypairsTestV21'),
                         '', [])]
        test_lists = [
            'compute.test_keypairs.KeypairsTestV210.test_create_keypair',
            'compute.test_keypairs.KeypairsTestV21.test_create_keypair',
            'compute.test_fake.FakeTest.test_fake_test']
        with mock.patch('stestr.selection.exclusion_reader',
                        return_value=exclude_list):
            result = selection.construct_list(test_lists,
                                              exclude_list='file',
                                              regexes=['fake_test'])
        self.assertEqual(
            list(result), ['compute.test_fake.FakeTest.test_fake_test'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""List the tests from a project and show them."""

from io import BytesIO
import os
import sys

from cliff import command

from stestr import config_file
from stestr import output


class List(command.Command):
    """List the tests for a project.

    You can use a filter just like with the run command to see exactly what
    tests match.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument("filters", nargs="*", default=None,
                            help="A list of string regex filters to initially "
                            "apply on the test list. Tests that match any of "
                            "the regexes will be used. (assuming any other "
                            "filtering specified also uses it)")
        parser.add_argument('--exclude-list', '-e',
                            default=None, dest='exclude_list',
                            help='Path to an exclusion list file, this file '
                                 'contains a separate regex exclude on each '
                                 'newline')
        parser.add_argument('--include-list', '-i',
                            default=None, dest='include_list',
                            help='Path to an inclusion list file, this file '
                                 'contains a separate regex on each newline.')
        parser.add_argument('--exclude-regex', '-E',
                            default=None, dest='exclude_regex',
                            help='Test rejection regex. If a test cases name '
                            'matches on re.search() operation , '
                            'it will be removed from the final test list. '
                            'Effectively the exclusion-regexp is added to '
                            'exclusion regexp list, but you do need to edit a '
                            'file. The exclusion filtering happens after the '
                            'initial safe list selection, which by default is '
                            'everything.')
        return parser

    def take_action(self, parsed_args):
        args = parsed_args
        filters = parsed_args.filters or None
        return list_command(config=self.app_args.config,
                            repo_url=self.app_args.repo_url,
                            group_regex=self.app_args.group_regex,
                            test_path=self.app_args.test_path,
                            top_dir=self.app_args.top_dir,
                            exclude_list=args.exclude_list,
                            include_list=args.include_list,
                            exclude_regex=args.exclude_regex,
                            filters=filters)


def list_command(config='.stestr.conf', repo_url=None,
                 test_path=None, top_dir=None, group_regex=None,
                 exclude_list=None,
                 include_list=None,
                 exclude_regex=None, filters=None,
                 stdout=sys.stdout):
    """Print a list of test_ids for a project

    This function will print the test_ids for tests in a project. You can
    filter the output just like with the run command to see exactly what
    will be run.

    :param str config: The path to the stestr config file. Must be a string.
    :param str repo_url: The url of the repository to use.
    :param str test_path: Set the test path to use for unittest discovery.
        If both this and the corresponding config file option are set, this
        value will be used.
    :param str top_dir: The top dir to use for unittest discovery. This takes
        precedence over the value in the config file. (if one is present in
        the config file)
    :param str group_regex: Set a group regex to use for grouping tests
        together in the stestr scheduler. If both this and the corresponding
        config file option are set this value will be used.
    :param str exclude_list: Path to an exclusion list file, this file
        contains a separate regex exclude on each newline.
    :param str include_list: Path to an inclusion list file, this file
        contains a separate regex on each newline.
    :param str exclude_regex: Test rejection regex. If a test cases name
        matches on re.search() operation, it will be removed from the final
        test list.
    :param list filters: A list of string regex filters to initially apply on
        the test list. Tests that match any of the regexes will be used.
        (assuming any other filtering specified also uses it)
    :param file stdout: The output file to write all output to. By default
        this is sys.stdout

    """
    ids = None
    if config and os.path.isfile(config):
        conf = config_file.TestrConf(config)
    elif os.path.isfile('tox.ini'):
        conf = config_file.TestrConf('tox.ini', section='stestr')
    else:
        conf = config_file.TestrConf(config)
    cmd = conf.get_run_command(
        regexes=filters,
        repo_url=repo_url, group_regex=group_regex,
        exclude_list=exclude_list,
        include_list=include_list,
        exclude_regex=exclude_regex,
        test_path=test_path, top_dir=top_dir)
    not_filtered = filters is None\
        and include_list is None and exclude_list is None\
        and exclude_regex is None
    try:
        cmd.setUp()
        # List tests if the fixture has not already needed to to filter.
        if not_filtered:
            ids = cmd.list_tests()
        else:
            ids = cmd.test_ids
        stream = BytesIO()
        for id in ids:
            stream.write(('%s\n' % id).encode('utf8'))
        stream.seek(0)
        output.output_stream(stream, output=stdout)
        return 0
    finally:
        cmd.cleanUp()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import sys

import voluptuous as vp
import yaml


def get_user_config(path=None):
    if not path:
        home_dir = os.path.expanduser("~")
        path = os.path.join(home_dir, '.stestr.yaml')
        if not os.path.isfile(path):
            path = os.path.join(os.path.join(home_dir, '.config'),
                                'stestr.yaml')
            if not os.path.isfile(path):
                path = None
        if not path:
            return None
    else:
        if not os.path.isfile(path):
            msg = 'The specified stestr user config is not a valid path'
            sys.exit(msg)

    return UserConfig(path)


class UserConfig:

    def __init__(self, path):
        self.schema = vp.Schema({
            vp.Optional('run'): {
                vp.Optional('concurrency'): int,
                vp.Optional('random'): bool,
                vp.Optional('no-subunit-trace'): bool,
                vp.Optional('color'): bool,
                vp.Optional('abbreviate'): bool,
                vp.Optional('slowest'): bool,
                vp.Optional('suppress-attachments'): bool,
                vp.Optional('all-attachments'): bool,
            },
            vp.Optional('failing'): {
                vp.Optional('list'): bool,
            },
            vp.Optional('last'): {
                vp.Optional('no-subunit-trace'): bool,
                vp.Optional('color'): bool,
                vp.Optional('suppress-attachments'): bool,
                vp.Optional('all-attachments'): bool,
            },
            vp.Optional('load'): {
                vp.Optional('force-init'): bool,
                vp.Optional('subunit-trace'): bool,
                vp.Optional('color'): bool,
                vp.Optional('abbreviate'): bool,
                vp.Optional('suppress-attachments'): bool,
                vp.Optional('all-attachments'): bool,
            },
            vp.Optional('history-list'): {
                vp.Optional('show-metadata'): bool,
            },
            vp.Optional('history-show'): {
                vp.Optional('no-subunit-trace'): bool,
                vp.Optional('color'): bool,
                vp.Optional('suppress-attachments'): bool,
                vp.Optional('all-attachments'): bool,
            }
        })
        with open(path) as fd:
            self.config = yaml.safe_load(fd.read())
        if self.config is None:
            self.config = {}
        try:
            self.schema(self.config)
        except vp.MultipleInvalid as e:
            msg = ('Provided user config file {} is invalid '
                   'because:\n{}'.format(path, str(e)))
            sys.exit(msg)

    @property
    def run(self):
        return self.config.get('run')

    @property
    def failing(self):
        return self.config.get('failing')

    @property
    def last(self):
        return self.config.get('last')

    @property
    def load(self):
        return self.config.get('load')

    @property
    def history_list(self):
        return self.config.get('history-list')

    @property
    def history_show(self):
        return self.config.get('history-show')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest

from stestr.subunit_runner import program
from stestr.tests import base


class TestGetById(base.TestCase):
    class CompatibleSuite(unittest.TestSuite):
        def filter_by_ids(self, test_ids):
            return unittest.TestSuite([item for item in self
                                       if item.id() in test_ids])

    @classmethod
    def setUpClass(cls):
        cls.test_suite = unittest.TestSuite([cls(name) for name in dir(cls)
                                             if name.startswith('test')])

    @classmethod
    def _test_name(cls, name):
        return f'{__name__}.{cls.__name__}.{name}'

    @staticmethod
    def _test_ids(suite):
        return [item.id() for item in suite]

    def test_filter_by_ids_case_no_tests(self):
        test_case = type(self)('test_filter_by_ids_case_no_tests')
        result = program.filter_by_ids(test_case, [
            'stestr.tests.test_load.TestLoadCommand.test_empty_with_pretty_out'
        ])
        self.assertEqual([], self._test_ids(result))

    def test_filter_by_ids_case_simple(self):
        name = 'test_filter_by_ids_case_simple'
        test_name = self._test_name(name)
        test_case = type(self)(name)
        result = program.filter_by_ids(test_case, [test_name])
        self.assertEqual([test_name], self._test_ids(result))

    def test_filter_by_ids_suite_no_tests(self):
        result = program.filter_by_ids(self.test_suite, [
            'stestr.tests.test_load.TestLoadCommand.test_empty_with_pretty_out'
        ])
        self.assertEqual([], self._test_ids(result))

    def test_filter_by_ids_suite_simple(self):
        test_name = self._test_name('test_filter_by_ids_suite_simple')
        result = program.filter_by_ids(self.test_suite, [test_name])
        self.assertEqual([test_name], self._test_ids(result))

    def test_filter_by_ids_suite_preserves_order(self):
        names = ('test_filter_by_ids_suite_simple',
                 'test_filter_by_ids_suite_preserves_order',
                 'test_filter_by_ids_suite_no_tests')
        test_names = [self._test_name(name) for name in names]
        result = program.filter_by_ids(self.test_suite, test_names)
        self.assertEqual(test_names, self._test_ids(result))

    def test_filter_by_ids_suite_no_duplicates(self):
        names = ('test_filter_by_ids_suite_simple',
                 'test_filter_by_ids_suite_preserves_order')
        expected = [self._test_name(name) for name in names]
        # Create duplicates in reversed order
        test_names = expected + expected[::-1]
        result = program.filter_by_ids(self.test_suite, test_names)
        self.assertEqual(expected, self._test_ids(result))

    def test_filter_by_ids_compatible_no_tests(self):
        test_suite = self.CompatibleSuite(self.test_suite)
        result = program.filter_by_ids(test_suite, [
            'stestr.tests.test_load.TestLoadCommand.test_empty_with_pretty_out'
        ])
        self.assertEqual([], self._test_ids(result))

    def test_filter_by_ids_compatible_simple(self):
        test_suite = self.CompatibleSuite(self.test_suite)
        test_name = self._test_name('test_filter_by_ids_compatible_simple')
        result = program.filter_by_ids(test_suite, [test_name])
        self.assertEqual([test_name], self._test_ids(result))

    def test_filter_by_ids_compatible_preserves_order(self):
        test_suite = self.CompatibleSuite(self.test_suite)
        names = ('test_filter_by_ids_compatible_simple',
                 'test_filter_by_ids_compatible_preserves_order',
                 'test_filter_by_ids_compatible_no_tests')
        test_names = [self._test_name(name) for name in names]
        result = program.filter_by_ids(test_suite, test_names)
        self.assertEqual(test_names, self._test_ids(result))

    def test_filter_by_ids_compatible_no_duplicates(self):
        test_suite = self.CompatibleSuite(self.test_suite)
        names = ('test_filter_by_ids_compatible_simple',
                 'test_filter_by_ids_compatible_preserves_order')
        expected = [self._test_name(name) for name in names]
        # Create duplicates in reversed order
        test_names = expected + expected[::-1]
        result = program.filter_by_ids(test_suite, test_names)
        self.assertEqual(expected, self._test_ids(result))

    def test_filter_by_ids_no_duplicates_preserve_order(self):
        test_suite = unittest.TestSuite(
            [self.CompatibleSuite(self.test_suite),
             self.test_suite,
             type(self)('test_filter_by_ids_no_duplicates_preserve_order')])
        names = ('test_filter_by_ids_compatible_simple',
                 'test_filter_by_ids_suite_simple',
                 'test_filter_by_ids_compatible_preserves_order')
        expected = [self._test_name(name) for name in names]
        # Create duplicates in reversed order
        test_names = expected + expected[::-1]
        result = program.filter_by_ids(test_suite, test_names)
        self.assertEqual(expected, self._test_ids(result))



# Copyright 2019 Matthew Treinish
# Copyright (c) 2009 testtools developers.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import os
import sys
import unittest

import extras


def filter_by_ids(suite_or_case, test_ids):
    """Return a test suite with the intersection of suite_or_case and test_ids.

    :param suite_or_case: A test suite or test case.
    :param test_ids: Something that supports the __contains__ protocol.
    :return: suite_or_case, unless suite_or_case was a case that itself
        fails the predicate when it will return a new unittest.TestSuite with
        no contents.

    For subclasses of TestSuite, filtering is done by:
        - attempting to call suite.filter_by_ids(test_ids)
        - if there is no method, iterating the suite and identifying tests to
          remove, then removing them from _tests, manually recursing into
          each entry.

    For objects with an id() method - TestCases, filtering is done by:
        - attempting to return case.filter_by_ids(test_ids)
        - if there is no such method, checking for case.id() in test_ids
          and returning case if it is, or TestSuite() if it is not.

    For anything else, it is not filtered - it is returned as-is.

    To provide compatibility with this routine for a custom TestSuite, just
    define a filter_by_ids() method that will return a TestSuite equivalent to
    the original minus any tests not in test_ids.
    Similarly to provide compatibility for a custom TestCase that does
    something unusual define filter_by_ids to return a new TestCase object
    that will only run test_ids that are in the provided container. If none
    would run, return an empty TestSuite().

    The contract for this function does not require mutation - each filtered
    object can choose to return a new object with the filtered tests. However
    because existing custom TestSuite classes in the wild do not have this
    method, we need a way to copy their state correctly which is tricky:
    thus the backwards-compatible code paths attempt to mutate in place rather
    than guessing how to reconstruct a new suite.
    """
    def _filter_tests(to_filter, id_map):
        # Compatible objects
        if extras.safe_hasattr(to_filter, 'filter_by_ids'):
            res = to_filter.filter_by_ids(test_ids)
            _filter_tests(res, id_map)
        # TestCase objects.
        elif extras.safe_hasattr(to_filter, 'id'):
            test_id = to_filter.id()
            if test_id in test_ids:
                id_map[test_id] = to_filter
        # Standard TestSuites or derived classes.
        elif isinstance(to_filter, unittest.TestSuite):
            for item in to_filter:
                _filter_tests(item, id_map)

    id_map = {}
    _filter_tests(suite_or_case, id_map)

    result = unittest.TestSuite()
    for test_id in test_ids:
        if test_id in id_map:
            # Use pop to avoid duplicates
            result.addTest(id_map.pop(test_id))

    return result


def iterate_tests(test_suite_or_case):
    """Iterate through all of the test cases in 'test_suite_or_case'."""
    try:
        suite = iter(test_suite_or_case)
    except TypeError:
        yield test_suite_or_case
    else:
        for test in suite:
            for subtest in iterate_tests(test):
                yield subtest


def list_test(test):
    """Return the test ids that would be run if test() was run.

    When things fail to import they can be represented as well, though
    we use an ugly hack (see http://bugs.python.org/issue19746 for details)
    to determine that. The difference matters because if a user is
    filtering tests to run on the returned ids, a failed import can reduce
    the visible tests but it can be impossible to tell that the selected
    test would have been one of the imported ones.

    :return: A tuple of test ids that would run and error strings
        describing things that failed to import.
    """
    unittest_import_strs = {
        'unittest2.loader.ModuleImportFailure.',
        'unittest.loader.ModuleImportFailure.',
        'discover.ModuleImportFailure.'
        }
    test_ids = []
    errors = []
    for test in iterate_tests(test):
        # Much ugly.
        for prefix in unittest_import_strs:
            if test.id().startswith(prefix):
                errors.append(test.id()[len(prefix):])
                break
        else:
            test_ids.append(test.id())
    return test_ids, errors


class TestProgram(unittest.TestProgram):

    # defaults for testing
    module = None
    verbosity = 1
    failfast = catchbreak = buffer = progName = None
    _discovery_parser = None

    def __init__(self, module='__main__', defaultTest=None, argv=None,
                 testRunner=None, testLoader=unittest.defaultTestLoader,
                 exit=False, verbosity=1, failfast=None, catchbreak=None,
                 buffer=None, warnings=None, tb_locals=False):
        if isinstance(module, str):
            self.module = __import__(module)
            for part in module.split('.')[1:]:
                self.module = getattr(self.module, part)
        else:
            self.module = module
        if argv is None:
            argv = sys.argv

        self.exit = exit
        self.failfast = failfast
        self.catchbreak = catchbreak
        self.verbosity = verbosity
        self.buffer = buffer
        self.tb_locals = tb_locals
        if warnings is None and not sys.warnoptions:
            # even if DeprecationWarnings are ignored by default
            # print them anyway unless other warnings settings are
            # specified by the warnings arg or the -W python flag
            self.warnings = 'default'
        else:
            # here self.warnings is set either to the value passed
            # to the warnings args or to None.
            # If the user didn't pass a value self.warnings will
            # be None. This means that the behavior is unchanged
            # and depends on the values passed to -W.
            self.warnings = warnings
        self.defaultTest = defaultTest
        # XXX: Local edit (see http://bugs.python.org/issue22860)
        self.listtests = False
        self.load_list = None
        self.testRunner = testRunner
        self.testLoader = testLoader
        self.progName = os.path.basename(argv[0])
        self.parseArgs(argv)
        # XXX: Local edit (see http://bugs.python.org/issue22860)
        if self.load_list:
            # TODO(mtreinish): preserve existing suites (like testresources
            # does in OptimisingTestSuite.add, but with a standard protocol).
            # This is needed because the load_tests hook allows arbitrary
            # suites, even if that is rarely used.
            source = open(self.load_list, 'rb')
            try:
                lines = source.readlines()
            finally:
                source.close()
            test_ids = [line.strip().decode('utf-8') for line in lines]
            self.test = filter_by_ids(self.test, test_ids)

        # XXX: Local edit (see http://bugs.python.org/issue22860)
        if not self.listtests:
            self.runTests()
        else:
            runner = self._get_runner()
            if extras.safe_hasattr(runner, 'list'):
                try:
                    runner.list(self.test, loader=self.testLoader)
                except TypeError:
                    runner.list(self.test)
            else:
                for test in iterate_tests(self.test):
                    sys.stdout.write('%s\n' % test.id())

    def _getParentArgParser(self):
        parser = super(TestProgram, self)._getParentArgParser()
        # XXX: Local edit (see http://bugs.python.org/issue22860)
        parser.add_argument(
            '-l', '--list', dest='listtests', default=False,
            action='store_true', help='List tests rather than executing them')
        parser.add_argument(
            '--load-list', dest='load_list', default=None,
            help='Specifies a file containing test ids, only tests matching '
                 'those ids are executed')
        return parser

    def _get_runner(self):
        testRunner = self.testRunner
        try:
            testRunner = self.testRunner(failfast=self.failfast,
                                         tb_locals=self.tb_locals)
        except TypeError:
            testRunner = self.testRunner()
        # If for some reason we failed to initialize the runner initialize
        # with defaults
        if isinstance(testRunner, functools.partial):
            testRunner = self.testRunner()
        return testRunner

    def runTests(self):
        if self.catchbreak:
            unittest.installHandler()
        testRunner = self._get_runner()
        self.result = testRunner.run(self.test)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import ddt

from stestr import config_file
from stestr.tests import base


@ddt.ddt
class TestTestrConf(base.TestCase):

    @mock.patch.object(config_file.configparser, 'ConfigParser')
    def setUp(self, mock_ConfigParser):
        super().setUp()
        self._testr_conf = config_file.TestrConf(mock.sentinel.config_file)
        self._testr_conf.parser = mock.Mock()

    @mock.patch.object(config_file.util, 'get_repo_open')
    @mock.patch.object(config_file.test_processor, 'TestProcessorFixture')
    @mock.patch.object(config_file, 'sys')
    @mock.patch('os.path.exists', new=lambda x: True)
    def _check_get_run_command(self, mock_sys, mock_TestProcessorFixture,
                               mock_get_repo_open, platform='win32',
                               group_regex='.*', parallel_class=False,
                               sys_executable='/usr/bin/python',
                               expected_python='/usr/bin/python',
                               expected_group_callback=mock.ANY,
                               environment=None):
        mock_sys.platform = platform
        mock_sys.executable = sys_executable
        if environment is None:
            environment = {'PYTHON': ''}

        with mock.patch.dict('os.environ', environment):
            fixture = \
                self._testr_conf.get_run_command(test_path='fake_test_path',
                                                 top_dir='fake_top_dir',
                                                 group_regex=group_regex,
                                                 parallel_class=parallel_class)

        self.assertEqual(mock_TestProcessorFixture.return_value, fixture)
        mock_get_repo_open.assert_called_once_with(repo_url=None)
        command = '"%s" -m stestr.subunit_runner.run discover -t "%s" "%s" ' \
                  '$LISTOPT $IDOPTION' % (expected_python, 'fake_top_dir',
                                          'fake_test_path')
        # Ensure TestProcessorFixture is created with defaults except for where
        # we specfied and with the correct python.
        mock_TestProcessorFixture.assert_called_once_with(
            None, command, "--list", "--load-list $IDFILE",
            mock_get_repo_open.return_value,
            exclude_regex=None,
            exclude_list=None, concurrency=0,
            group_callback=expected_group_callback,
            test_filters=None, randomize=False, serial=False,
            include_list=None, worker_path=None)

    @mock.patch.object(config_file, 'sys')
    def _check_get_run_command_exception(self, mock_sys, platform='win32',
                                         sys_executable='/usr/bin/python',
                                         environment=None):
        mock_sys.platform = platform
        mock_sys.executable = sys_executable
        if environment is None:
            environment = {'PYTHON': ''}

        with mock.patch.dict('os.environ', environment):
            self.assertRaises(RuntimeError, self._testr_conf.get_run_command,
                              test_path='fake_test_path',
                              top_dir='fake_top_dir')

    def test_get_run_command_linux(self):
        self._check_get_run_command(
            platform='linux2',
            expected_python='/usr/bin/python')

    def test_get_run_command_emptysysexecutable_noenv(self):
        self._check_get_run_command_exception(
            platform='linux2',
            sys_executable=None)

    def test_get_run_command_emptysysexecutable_win32(self):
        self._check_get_run_command_exception(
            platform='win32', sys_executable=None,
            environment={'PYTHON': 'python3'})

    def test_get_run_command_emptysysexecutable_withenv(self):
        self._check_get_run_command(
            platform='linux2', sys_executable=None,
            expected_python='${PYTHON}',
            environment={'PYTHON': '/usr/bin/python3'})

    def test_get_run_command_win32(self):
        self._check_get_run_command()

    def test_get_run_command_parallel_class(self):
        self._check_get_run_command(parallel_class=True)

    def test_get_run_command_nogroup_regex_noparallel_class(self):
        self._testr_conf.parser.has_option.return_value = False
        self._check_get_run_command(group_regex='',
                                    expected_group_callback=None)

    @ddt.data(('.\\', '.\\\\'),
              ('a\\b\\', 'a\\b\\\\'),
              ('a\\b', 'a\\b'))
    @ddt.unpack
    @mock.patch('os.sep', new='\\')
    def test_sanitize_dir_win32(self, path, expected):
        sanitized = self._testr_conf._sanitize_path(path)
        self.assertEqual(expected, sanitized)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Show the longest running tests in the repository."""

import math
from operator import itemgetter
import sys

from cliff import command

from stestr import output
from stestr.repository import util


class Slowest(command.Command):
    """Show the slowest tests from the last test run.

    This command shows a table, with the longest running tests at the top.
    """

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--all", action="store_true",
            default=False, help="Show timing for all tests.")
        return parser

    def take_action(self, parsed_args):
        args = parsed_args
        return slowest(repo_url=self.app_args.repo_url,
                       show_all=args.all)


def format_times(times):
    times = list(times)
    precision = 3
    digits_before_point = 1
    for time in times:
        if time[1] <= 0:
            continue
        digits_before_point = int(math.log10(time[1])) + 1
        break
    min_length = digits_before_point + precision + 1

    def format_time(time):
        # Limit the number of digits after the decimal
        # place, and also enforce a minimum width
        # based on the longest duration
        return "%*.*f" % (min_length, precision, time)
    times = [(name, format_time(time)) for name, time in times]
    return times


def slowest(repo_url=None, show_all=False,
            stdout=sys.stdout):
    """Print the slowest times from the last run in the repository

    This function will print to STDOUT the 10 slowests tests in the last run.
    Optionally, using the ``show_all`` argument, it will print all the tests,
    instead of just 10. sorted by time.

    :param str repo_url: The url of the repository to use.
    :param bool show_all: Show timing for all tests.
    :param file stdout: The output file to write all output to. By default
        this is sys.stdout

    :return return_code: The exit code for the command. 0 for success and > 0
        for failures.
    :rtype: int
    """

    repo = util.get_repo_open(repo_url=repo_url)
    try:
        latest_id = repo.latest_id()
    except KeyError:
        return 3
    # what happens when there is no timing info?
    test_times = repo.get_test_times(repo.get_test_ids(latest_id))
    known_times = list(test_times['known'].items())
    known_times.sort(key=itemgetter(1), reverse=True)
    if len(known_times) > 0:
        # By default show 10 rows
        if not show_all:
            known_times = known_times[:10]
        known_times = format_times(known_times)
        header = ('Test id', 'Runtime (s)')
        rows = [header] + known_times
        output.output_table(rows, output=stdout)
    return 0



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from cliff import app
from cliff import commandmanager
from stestr import version

__version__ = version.version_info.version_string_with_vcs()


class StestrCLI(app.App):

    def __init__(self):
        super().__init__(
            description="A parallel Python test runner built around subunit",
            version=__version__,
            command_manager=commandmanager.CommandManager('stestr.cm'),
            deferred_help=True,
            )

    def initialize_app(self, argv):
        self.options.debug = True
        self.LOG.debug('initialize_app')

    def prepare_to_run_command(self, cmd):
        self.LOG.debug('prepare_to_run_command %s', cmd.__class__.__name__)
        group_regex = r'([^\.]*\.)*' \
            if cmd.app_args.parallel_class else cmd.app_args.group_regex
        cmd.app_args.group_regex = group_regex

    def clean_up(self, cmd, result, err):
        self.LOG.debug('clean_up %s', cmd.__class__.__name__)
        if err:
            self.LOG.debug('got an error: %s', err)

    def build_option_parser(self, description, version, argparse_kwargs=None):
        parser = super().build_option_parser(description, version,
                                             argparse_kwargs)
        parser = self._set_common_opts(parser)
        return parser

    def _set_common_opts(self, parser):
        parser.add_argument('--user-config', dest='user_config', default=None,
                            help='An optional path to a default user config '
                                 'file if one is not specified ~/.stestr.yaml '
                                 'and ~/.config/stestr.yaml will be tried in '
                                 'that order')
        parser.add_argument('-d', '--here', dest='here',
                            help="Set the directory or url that a command "
                                 "should run from. This affects all default "
                                 "path lookups but does not affect paths "
                                 "supplied to the command.",
                            default=None, type=str)
        parser.add_argument('--config', '-c', dest='config',
                            default='.stestr.conf',
                            help="Set a stestr config file to use with this "
                                 "command. If one isn't specified then "
                                 ".stestr.conf in the directory that a command"
                                 " is running from is used")
        parser.add_argument('--repo-url', '-u', dest='repo_url',
                            default=None,
                            help="Set the repo url to use. This should just "
                                 "be the path to the stestr repository "
                                 "directory")
        parser.add_argument('--test-path', '-t', dest='test_path',
                            default=None,
                            help="Set the test path to use for unittest "
                                 "discovery. If both this and the "
                                 "corresponding config file option are set, "
                                 "this value will be used.")
        parser.add_argument('--top-dir', dest='top_dir',
                            default=None,
                            help="Set the top dir to use for unittest "
                                 "discovery. If both this and the "
                                 "corresponding config file option are set, "
                                 "this value will be used.")
        parser.add_argument('--group-regex', '--group_regex', '-g',
                            dest='group_regex',
                            default=None,
                            help="Set a group regex to use for grouping tests"
                                 " together in the stestr scheduler. If "
                                 "both this and the corresponding config file "
                                 "option are set this value will be used.")
        parser.add_argument('--parallel-class', '-p',
                            action='store_true',
                            default=False,
                            help="Set the flag to group tests by class. NOTE: "
                                 "This flag takes priority over the "
                                 "`--group-regex` option even if it's set.")

        return parser


def main(argv=sys.argv[1:]):
    cli = StestrCLI()
    return cli.run(argv)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from stestr.commands.failing import failing as failing_command
from stestr.commands.history import history_list as history_list_command
from stestr.commands.history import history_remove as history_remove_command
from stestr.commands.history import history_show as history_show_command
from stestr.commands.init import init as init_command
from stestr.commands.last import last as last_command
from stestr.commands.list import list_command
from stestr.commands.load import load as load_command
from stestr.commands.run import run_command
from stestr.commands.slowest import slowest as slowest_command


__all__ = ['failing_command', 'init_command', 'last_command',
           'list_command', 'load_command', 'run_command', 'slowest_command',
           'history_show_command', 'history_list_command',
           'history_remove_command']



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import re
import sys

import configparser

from stestr.repository import util
from stestr import test_processor


class TestrConf:
    """Create a TestrConf object to represent a specified config file

    This class is used to represent an stestr config file. Or in the case
    of a tox.ini file the stestr section in the tox.ini file

    :param str config_file: The path to the config file to use
    :param str section: The section to use for the stestr config. By default
        this is DEFATULT.
    """

    _escape_trailing_backslash_re = re.compile(r'(?<=[^\\])\\$')

    def __init__(self, config_file, section='DEFAULT'):
        self.parser = configparser.ConfigParser()
        self.parser.read(config_file)
        self.config_file = config_file
        self.section = section

    def _sanitize_path(self, path):
        if os.sep == '\\':
            # Trailing backslashes have to be escaped. Othwerise, the
            # command we're issuing will be incorrectly interpreted on
            # Windows.
            path = self._escape_trailing_backslash_re.sub(r'\\\\', path)
        return path

    def get_run_command(self, test_ids=None, regexes=None,
                        test_path=None, top_dir=None, group_regex=None,
                        repo_url=None,
                        serial=False, worker_path=None,
                        concurrency=0,
                        exclude_list=None,
                        include_list=None,
                        exclude_regex=None,
                        randomize=False, parallel_class=None):
        """Get a test_processor.TestProcessorFixture for this config file

        Any parameters about running tests will be used for initialize the
        output fixture so the settings are correct when that fixture is used
        to run tests. Parameters will take precedence over values in the config
        file.

        :param options: A argparse Namespace object of the cli options that
            were used in the invocation of the original CLI command that
            needs a TestProcessorFixture
        :param list test_ids: an optional list of test_ids to use when running
            tests
        :param list regexes: an optional list of regex strings to use for
            filtering the tests to run. See the test_filters parameter in
            TestProcessorFixture to see how this is used.
        :param str test_path: Set the test path to use for unittest discovery.
            If both this and the corresponding config file option are set, this
            value will be used.
        :param str top_dir: The top dir to use for unittest discovery. This
            takes precedence over the value in the config file. (if one is
            present in the config file)
        :param str group_regex: Set a group regex to use for grouping tests
            together in the stestr scheduler. If both this and the
            corresponding config file option are set this value will be used.
        :param str repo_url: The url of the repository to use.
        :param bool serial: If tests are run from the returned fixture, they
            will be run serially
        :param str worker_path: Optional path of a manual worker grouping file
            to use for the run.
        :param int concurrency: How many processes to use. The default (0)
            autodetects your CPU count and uses that.
        :param str exclude_list: Path to an exclusion list file, this
            file contains a separate regex exclude on each newline.
        :param str include_list: Path to an inclusion list file, this
            file contains a separate regex on each newline.
        :param str exclude_regex: Test rejection regex. If a test cases name
            matches on re.search() operation, it will be removed from the final
            test list.
        :param bool randomize: Randomize the test order after they are
            partitioned into separate workers
        :param bool parallel_class: Set the flag to group tests together in the
            stestr scheduler by class. If both this and the corresponding
            config file option which includes `group-regex` are set, this value
            will be used.

        :returns: a TestProcessorFixture object for the specified config file
            and any arguments passed into this function
        :rtype: test_processor.TestProcessorFixture
        """

        if not test_path and self.parser.has_option(self.section, 'test_path'):
            test_path = self.parser.get(self.section, 'test_path')
        elif not test_path:
            sys.exit("No test_path can be found in either the command line "
                     "options nor in the specified config file {}.  Please "
                     "specify a test path either in the config file or via "
                     "the --test-path argument".format(self.config_file))
        if not top_dir and self.parser.has_option(self.section, 'top_dir'):
            top_dir = self.parser.get(self.section, 'top_dir')
        elif not top_dir:
            top_dir = './'

        test_path = self._sanitize_path(test_path)
        top_dir = self._sanitize_path(top_dir)

        stestr_python = sys.executable
        # let's try to be explicit, even if it means a longer set of ifs
        if sys.platform == 'win32':
            # it may happen, albeit rarely
            if not stestr_python:
                raise RuntimeError("The Python interpreter was not found")
            python = stestr_python
        else:
            if os.environ.get('PYTHON'):
                python = '${PYTHON}'
            elif stestr_python:
                python = stestr_python
            else:
                raise RuntimeError("The Python interpreter was not found and "
                                   "PYTHON is not set")

        # The python binary path may contain whitespaces.
        if os.path.exists('"%s"' % python):
            python = '"%s"' % python

        command = '%s -m stestr.subunit_runner.run discover -t "%s" "%s" ' \
                  '$LISTOPT $IDOPTION' % (python, top_dir, test_path)
        listopt = "--list"
        idoption = "--load-list $IDFILE"
        # If the command contains $IDOPTION read that command from config
        # Use a group regex if one is defined
        if parallel_class:
            group_regex = r'([^\.]*\.)*'
        if not group_regex \
                and self.parser.has_option(self.section, 'parallel_class') \
                and self.parser.getboolean(self.section, 'parallel_class'):
            group_regex = r'([^\.]*\.)*'
        if not group_regex and self.parser.has_option(self.section,
                                                      'group_regex'):
            group_regex = self.parser.get(self.section, 'group_regex')
        if group_regex:
            def group_callback(test_id, regex=re.compile(group_regex)):
                match = regex.match(test_id)
                if match:
                    return match.group(0)
        else:
            group_callback = None
        # Handle the results repository
        repository = util.get_repo_open(repo_url=repo_url)
        return test_processor.TestProcessorFixture(
            test_ids, command, listopt, idoption, repository,
            test_filters=regexes, group_callback=group_callback, serial=serial,
            worker_path=worker_path, concurrency=concurrency,
            exclude_list=exclude_list,
            exclude_regex=exclude_regex,
            include_list=include_list, randomize=randomize)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import sys

import subunit
import testtools


def output_table(table, output=sys.stdout):
    """Display a table of information.

    :param table: A list of sets representing each row in the table. Each
        element in the set represents a column in the table.
    :param output: The output file object to write the table to. By default
        this is sys.stdout
    """
    # stringify
    contents = []
    for row in table:
        new_row = []
        for column in row:
            new_row.append(str(column))
        contents.append(new_row)
    if not contents:
        return
    widths = [0] * len(contents[0])
    for row in contents:
        for idx, column in enumerate(row):
            if widths[idx] < len(column):
                widths[idx] = len(column)
    # Show a row
    outputs = []

    def show_row(row):
        for idx, column in enumerate(row):
            outputs.append(column)
            if idx == len(row) - 1:
                outputs.append('\n')
                return
            # spacers for the next column
            outputs.append(' ' * (widths[idx] - len(column)))
            outputs.append('  ')

    show_row(contents[0])
    # title spacer
    for idx, width in enumerate(widths):
        outputs.append('-' * width)
        if idx == len(widths) - 1:
            outputs.append('\n')
            continue
        outputs.append('  ')
    for row in contents[1:]:
        show_row(row)
    output.write(''.join(outputs))


def output_tests(tests, output=sys.stdout):
    """Display a list of tests.

    :param tests: A list of test objects to output
    :param output: The output file object to write the list to. By default
        this is sys.stdout
    """

    for test in tests:
        id_str = test.id()
        output.write(str(id_str))
        output.write('\n')


def make_result(get_id, output=sys.stdout):
    serializer = subunit.StreamResultToBytes(output)
    # By pass user transforms - just forward it all,
    result = serializer
    # and interpret everything as success.
    summary = testtools.StreamSummary()
    summary.startTestRun()
    summary.stopTestRun()
    return result, summary


def output_summary(successful, tests, tests_delta, time, time_delta, values,
                   output=sys.stdout):
    """Display a summary view for the test run.

    :param bool successful: Was the test run successful
    :param int tests: The number of tests that ran
    :param int tests_delta: The change in the number of tests that ran since
        the last run
    :param float time: The number of seconds that it took for the run to
        execute
    :param float time_delta: The change in run time since the last run
    :param values: A list of sets that are used for a breakdown of statuses
        other than success. Each set is in the format:
        (status, number of tests, change in number of tests).
    :param output: The output file object to use. This defaults to stdout
    """
    summary = []
    a = summary.append
    if tests:
        a("Ran {}".format(tests))
        if tests_delta:
            a(" (%+d)" % (tests_delta,))
        a(" tests")
    if time:
        if not summary:
            a("Ran tests")
        a(" in {:0.3f}s".format(time))
        if time_delta:
            a(" ({:+0.3f}s)".format(time_delta))
    if summary:
        a("\n")
    if successful:
        a('PASSED')
    else:
        a('FAILED')
    if values:
        a(' (')
        values_strings = []
        for name, value, delta in values:
            value_str = '{}={}'.format(name, value)
            if delta:
                value_str += ' (%+d)' % (delta,)
            values_strings.append(value_str)
        a(', '.join(values_strings))
        a(')')
    output.write(''.join(summary) + '\n')


def output_stream(stream, output=sys.stdout):
    _binary_stdout = subunit.make_stream_binary(output)
    contents = stream.read(65536)
    assert type(contents) is bytes, \
        "Bad stream contents %r" % type(contents)
    # If there are unflushed bytes in the text wrapper, we need to sync..
    output.flush()
    while contents:
        _binary_stdout.write(contents)
        contents = stream.read(65536)
    _binary_stdout.flush()


class ReturnCodeToSubunit:
    """Converts a process return code to a subunit error on the process stdout.

    The ReturnCodeToSubunit object behaves as a read-only stream, supplying
    the read, readline and readlines methods. If the process exits non-zero a
    synthetic test is added to the output, making the error accessible to
    subunit stream consumers. If the process closes its stdout and then does
    not terminate, reading from the ReturnCodeToSubunit stream will hang.

    :param process: A subprocess.Popen object that is
        generating subunit.
    """

    def __init__(self, process):
        """Adapt a process to a readable stream."""
        self.proc = process
        self.done = False
        self.source = self.proc.stdout
        self.lastoutput = bytes((b'\n')[0])

    def __del__(self):
        self.proc.wait()

    def _append_return_code_as_test(self):
        if self.done is True:
            return
        self.source = io.BytesIO()
        returncode = self.proc.wait()
        if returncode != 0:
            if self.lastoutput != bytes((b'\n')[0]):
                # Subunit V1 is line orientated, it has to start on a fresh
                # line. V2 needs to start on any fresh utf8 character border
                # - which is not guaranteed in an arbitrary stream endpoint, so
                # injecting a \n gives us such a guarantee.
                self.source.write(bytes('\n'))
            stream = subunit.StreamResultToBytes(self.source)
            stream.status(test_id='process-returncode', test_status='fail',
                          file_name='traceback',
                          mime_type='text/plain;charset=utf8',
                          file_bytes=(
                              'returncode %d' % returncode).encode('utf8'))
        self.source.seek(0)
        self.done = True

    def read(self, count=-1):
        if count == 0:
            return ''
        result = self.source.read(count)
        if result:
            self.lastoutput = result[-1]
            return result
        self._append_return_code_as_test()
        return self.source.read(count)

    def readline(self):
        result = self.source.readline()
        if result:
            self.lastoutput = result[-1]
            return result
        self._append_return_code_as_test()
        return self.source.readline()

    def readlines(self):
        result = self.source.readlines()
        if result:
            self.lastoutput = result[-1][-1]
        self._append_return_code_as_test()
        result.extend(self.source.readlines())
        return result



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
from unittest import mock

from stestr.commands import run
from stestr.tests import base


class TestRunCommand(base.TestCase):
    def test_to_int_positive_int(self):
        self.assertEqual(29, run._to_int(29))

    def test_to_int_positive_int_str(self):
        self.assertEqual(42, run._to_int('42'))

    def test_to_int_negative_int(self):
        self.assertEqual(-2, run._to_int(-2))

    def test_to_int_negative_int_str(self):
        self.assertEqual(-45, run._to_int('-45'))

    def test_to_int_invalid_str(self):
        fake_stderr = io.StringIO()
        out = run._to_int('I am not an int', out=fake_stderr)
        expected = (
            'Unable to convert "I am not an int" to an integer.  '
            'Using 0.\n')
        self.assertEqual(fake_stderr.getvalue(), expected)
        self.assertEqual(0, out)

    def test_to_int_none(self):
        fake_stderr = io.StringIO()
        out = run._to_int(None, out=fake_stderr)
        expected = (
            'Unable to convert "None" to an integer.  '
            'Using 0.\n')
        self.assertEqual(fake_stderr.getvalue(), expected)
        self.assertEqual(0, out)


class TestLoadListAndFilter(base.TestCase):
    @mock.patch('builtins.open', side_effect=FileNotFoundError)
    def test__load_list_and_filter_file_not_found(self, mock_open):
        filename = 'my_filename'
        self.assertRaises(FileNotFoundError,
                          run._load_list_and_filter, filename, None)

        mock_open.assert_called_once_with(filename, 'rb')
        mock_open.return_value.__enter__.return_value.read.assert_not_called()

    @mock.patch('builtins.open')
    def test__load_list_and_filter_empty_file(self, mock_open):
        mock_read = mock_open.return_value.__enter__.return_value.read
        mock_read.return_value = b''
        filename = 'my_filename'
        res = run._load_list_and_filter(filename, ['mytest'])
        self.assertListEqual([], res)
        mock_open.assert_called_once_with(filename, 'rb')
        mock_read.assert_called_once_with()

    @mock.patch('builtins.open')
    def test__load_list_and_filter_no_filter(self, mock_open):
        mock_read = mock_open.return_value.__enter__.return_value.read
        tests = ['test1', 'tests2', 'test3']
        mock_read.return_value = '\n'.join(tests).encode()
        filename = 'my_filename'

        res = run._load_list_and_filter(filename, None)

        self.assertListEqual(tests, res)
        mock_open.assert_called_once_with(filename, 'rb')
        mock_read.assert_called_once_with()

    @mock.patch('builtins.open')
    def test__load_list_and_filter_filter(self, mock_open):
        mock_read = mock_open.return_value.__enter__.return_value.read
        tests = ['test1', 'tests2', 'test3']
        mock_read.return_value = '\n'.join(tests).encode()
        filename = 'my_filename'

        # Failed tests in different order from list and with additional ids to
        # confirm that order from list is preserved and only elements from list
        # are returned
        failed_tests = ['test4', 'test3', 'test1']
        expected = ['test1', 'test3']
        res = run._load_list_and_filter(filename, failed_tests)

        self.assertListEqual(expected, res)
        mock_open.assert_called_once_with(filename, 'rb')
        mock_read.assert_called_once_with()



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import io
import os
import re
import shutil
import subprocess
import tempfile

import fixtures
import subunit as subunit_lib
import testtools

from stestr.commands import list as list_cmd
from stestr.commands import run
from stestr.tests import base


class TestReturnCodes(base.TestCase):
    def setUp(self):
        super().setUp()
        # Setup test dirs
        self.directory = tempfile.mkdtemp(prefix='stestr-unit')
        self.addCleanup(shutil.rmtree, self.directory, ignore_errors=True)
        self.test_dir = os.path.join(self.directory, 'tests')
        os.mkdir(self.test_dir)
        # Setup Test files
        self.repo_root = os.path.abspath(os.curdir)
        self.testr_conf_file = os.path.join(self.directory, '.stestr.conf')
        self.setup_cfg_file = os.path.join(self.directory, 'setup.cfg')
        self.passing_file = os.path.join(self.test_dir, 'test_passing.py')
        self.failing_file = os.path.join(self.test_dir, 'test_failing.py')
        self.init_file = os.path.join(self.test_dir, '__init__.py')
        self.setup_py = os.path.join(self.directory, 'setup.py')
        self.user_config = os.path.join(self.directory, 'stestr.yaml')
        shutil.copy('stestr/tests/files/testr-conf', self.testr_conf_file)
        shutil.copy('stestr/tests/files/passing-tests', self.passing_file)
        shutil.copy('stestr/tests/files/failing-tests', self.failing_file)
        shutil.copy('setup.py', self.setup_py)
        shutil.copy('stestr/tests/files/setup.cfg', self.setup_cfg_file)
        shutil.copy('stestr/tests/files/__init__.py', self.init_file)
        shutil.copy('stestr/tests/files/stestr.yaml', self.user_config)

        self.stdout = io.StringIO()
        self.stderr = io.StringIO()
        # Change directory, run wrapper and check result
        self.addCleanup(os.chdir, self.repo_root)
        os.chdir(self.directory)
        subprocess.call('stestr init', shell=True)

    def _check_subunit(self, output_stream):
        stream = subunit_lib.ByteStreamToStreamResult(output_stream)
        starts = testtools.StreamResult()
        summary = testtools.StreamSummary()
        tests = []

        def _add_dict(test):
            tests.append(test)

        outcomes = testtools.StreamToDict(functools.partial(_add_dict))
        result = testtools.CopyStreamResult([starts, outcomes, summary])
        result.startTestRun()
        try:
            stream.run(result)
        finally:
            result.stopTestRun()
        self.assertThat(len(tests), testtools.matchers.GreaterThan(0))

    def assertRunExit(self, cmd, expected, subunit=False, stdin=None):
        if stdin:
            p = subprocess.Popen(
                "%s" % cmd, shell=True, stdin=subprocess.PIPE,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate(stdin)
        else:
            p = subprocess.Popen(
                "%s" % cmd, shell=True,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate()

        if not subunit:
            self.assertEqual(
                p.returncode, expected,
                "Stdout: {}; Stderr: {}".format(out, err))
            return (out, err)
        else:
            self.assertEqual(p.returncode, expected,
                             "Expected return code: %s doesn't match actual "
                             "return code of: %s" % (expected, p.returncode))
            output_stream = io.BytesIO(out)
            stream = subunit_lib.ByteStreamToStreamResult(output_stream)
            starts = testtools.StreamResult()
            summary = testtools.StreamSummary()
            tests = []

            def _add_dict(test):
                tests.append(test)

            outcomes = testtools.StreamToDict(functools.partial(_add_dict))
            result = testtools.CopyStreamResult([starts, outcomes, summary])
            result.startTestRun()
            try:
                stream.run(result)
            finally:
                result.stopTestRun()
            self.assertThat(len(tests), testtools.matchers.GreaterThan(0))
            return (out, err)

    def test_parallel_passing(self):
        self.assertRunExit('stestr run passing', 0)

    def test_parallel_passing_bad_regex(self):
        self.assertRunExit('stestr run bad.regex.foobar', 1)

    def test_parallel_fails(self):
        self.assertRunExit('stestr run', 1)

    def test_parallel_passing_xfail(self):
        self.assertRunExit('stestr run xfail', 0)

    def test_parallel_fails_unxsuccess(self):
        self.assertRunExit('stestr run unexpected', 1)

    def test_parallel_exclusion_list(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        with os.fdopen(fd, 'w') as exclusion_list:
            exclusion_list.write('fail')
        cmd = 'stestr run --exclude-list %s' % path
        self.assertRunExit(cmd, 0)

    def test_parallel_inclusion_list(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        with os.fdopen(fd, 'w') as inclusion_list:
            inclusion_list.write('passing')
        cmd = 'stestr run --include-list %s' % path
        self.assertRunExit(cmd, 0)

    def test_serial_passing(self):
        self.assertRunExit('stestr run --serial passing', 0)

    def test_serial_fails(self):
        self.assertRunExit('stestr run --serial', 1)

    def test_serial_exclusion_list(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        with os.fdopen(fd, 'w') as exclusion_list:
            exclusion_list.write('fail')
        cmd = 'stestr run --serial --exclude-list %s' % path
        self.assertRunExit(cmd, 0)

    def test_serial_inclusion_list(self):
        fd, path = tempfile.mkstemp()
        self.addCleanup(os.remove, path)
        with os.fdopen(fd, 'w') as inclusion_list:
            inclusion_list.write('passing')
        cmd = 'stestr run --serial --include-list %s' % path
        self.assertRunExit(cmd, 0)

    def test_serial_subunit_passing(self):
        self.assertRunExit('stestr --user-config stestr.yaml run --subunit '
                           '--serial passing', 0, subunit=True)

    def test_serial_subunit_failing(self):
        self.assertRunExit('stestr --user-config stestr.yaml run --subunit '
                           '--serial failing', 0, subunit=True)

    def test_parallel_subunit_passing(self):
        self.assertRunExit('stestr --user-config stestr.yaml run --subunit '
                           'passing', 0, subunit=True)

    def test_parallel_subunit_failing(self):
        self.assertRunExit('stestr --user-config stestr.yaml run --subunit '
                           'failing', 0, subunit=True)

    def test_slowest_passing(self):
        self.assertRunExit('stestr run --slowest passing', 0)

    def test_slowest_failing(self):
        self.assertRunExit('stestr run --slowest failing', 1)

    def test_until_failure_fails(self):
        self.assertRunExit('stestr run --until-failure', 1)

    def test_until_failure_with_subunit_fails(self):
        self.assertRunExit('stestr --user-config stestr.yaml run '
                           '--until-failure --subunit', 1, subunit=True)

    def test_with_parallel_class(self):
        # NOTE(masayukig): Ideally, it's better to figure out the
        # difference between with --parallel-class and without
        # --parallel-class. However, it's difficult to make such a
        # test from a command line based test.
        self.assertRunExit('stestr --parallel-class run passing', 0)

    def test_no_repo_dir(self):
        stestr_repo_dir = os.path.join(self.directory, '.stestr')
        shutil.rmtree(stestr_repo_dir, ignore_errors=True)
        # We can use stestr run even if there's no repo directory.
        self.assertRunExit('stestr run passing', 0)

    def test_empty_repo_dir(self):
        stestr_repo_dir = os.path.join(self.directory, '.stestr')
        shutil.rmtree(stestr_repo_dir, ignore_errors=True)
        os.mkdir(stestr_repo_dir)
        # We can initialize an empty repo directory.
        self.assertRunExit('stestr run passing', 0)

    def test_non_empty_repo_dir(self):
        stestr_repo_dir = os.path.join(self.directory, '.stestr')
        shutil.rmtree(stestr_repo_dir, ignore_errors=True)
        os.mkdir(stestr_repo_dir)
        with open(os.path.join(stestr_repo_dir, 'foo'), 'wt') as stream:
            stream.write('1\n')
        # We can't initialize a non-empty repo directory.
        self.assertRunExit('stestr run passing', 1)

    def test_list(self):
        self.assertRunExit('stestr list', 0)

    def _get_cmd_stdout(self, cmd):
        p = subprocess.Popen(cmd, shell=True,
                             stdout=subprocess.PIPE)
        out = p.communicate()
        self.assertEqual(0, p.returncode)
        return out

    def test_combine_results(self):
        self.assertRunExit('stestr run passing', 0)
        stdout = self._get_cmd_stdout(
            'stestr last --no-subunit-trace')
        stdout = str(stdout[0])
        test_count_split = stdout.split(' ')
        test_count = test_count_split[1]
        test_count = int(test_count)
        id_regex = re.compile(r'\(id=(.*?)\)')
        test_id = id_regex.search(stdout).group(0)
        self.assertRunExit('stestr run --combine passing', 0)
        combine_stdout = self._get_cmd_stdout(
            'stestr last --no-subunit-trace')[0]
        combine_stdout = str(combine_stdout)
        combine_test_count_split = combine_stdout.split(' ')
        combine_test_count = combine_test_count_split[1]
        combine_test_count = int(combine_test_count)
        combine_test_id = id_regex.search(combine_stdout).group(0)
        self.assertEqual(test_id, combine_test_id)

        # The test results from running the same tests twice with combine
        # should return a test count 2x as big at the end of the run
        self.assertEqual(test_count * 2, combine_test_count)

    def test_load_from_stdin(self):
        self.assertRunExit('stestr run passing', 0)
        stream = self._get_cmd_stdout(
            'stestr last --subunit')[0]
        self.assertRunExit('stestr load', 0, stdin=stream)

    def test_load_force_init(self):
        self.assertRunExit('stestr run passing', 0)
        stream = self._get_cmd_stdout(
            'stestr last --subunit')[0]
        # NOTE: --force-init should work here because there is an properly
        # initialized repository.
        self.assertRunExit('stestr load --force-init', 0, stdin=stream)

    def test_load_force_init_invalid(self):
        self.assertRunExit('stestr run passing', 0)
        stream = self._get_cmd_stdout(
            'stestr last --subunit')[0]
        os.remove(os.path.join(self.directory, '.stestr', 'format'))
        # NOTE: --force-init should fail here because there is an invalid
        # repository.
        self.assertRunExit('stestr load --force-init', 1, stdin=stream)

    def test_load_from_stdin_quiet(self):
        out, err = self.assertRunExit('stestr --user-config stestr.yaml -q '
                                      'run passing', 0)
        self.assertEqual(out.decode('utf-8'), '')
        # FIXME(masayukig): We get some warnings when we run a coverage job.
        # So, just ignore 'err' here.
        stream = self._get_cmd_stdout('stestr last --subunit')[0]
        out, err = self.assertRunExit('stestr --user-config stestr.yaml -q '
                                      'load', 0, stdin=stream)
        self.assertEqual(out.decode('utf-8'), '')
        self.assertEqual(err.decode('utf-8'), '')

    def test_history_list(self):
        self.assertRunExit('stestr run passing', 0)
        self.assertRunExit('stestr run', 1)
        self.assertRunExit('stestr run passing', 0)
        table = self.assertRunExit(
            'stestr history list', 0)[0].decode('utf8')
        self.assertIn("| 0      | True   |", table.split('\n')[3].rstrip())
        self.assertIn("| 1      | False  |", table.split('\n')[4].rstrip())
        self.assertIn("| 2      | True   |", table.split('\n')[5].rstrip())
        expected = """
+--------+--------+-----------+----------------------------------+
| Run ID | Passed | Runtime   | Date                             |
+--------+--------+-----------+----------------------------------+
""".rstrip()
        self.assertEqual(expected.strip(), '\n'.join(
            [x.rstrip() for x in table.split('\n')[:3]]).strip())

    def test_history_empty(self):
        table = self.assertRunExit(
            'stestr history list', 0)[0].decode('utf8')
        self.assertEqual("",
                         '\n'.join(
                             [x.rstrip() for x in table.split('\n')]).strip())

    def test_history_show_passing(self):
        self.assertRunExit('stestr run passing', 0)
        self.assertRunExit('stestr run', 1)
        self.assertRunExit('stestr run passing', 0)
        output, _ = self.assertRunExit('stestr history show 0', 0)
        lines = [x.rstrip() for x in output.decode('utf8').split('\n')]
        self.assertIn(' - Passed: 2', lines)
        self.assertIn(' - Failed: 0', lines)
        self.assertIn(' - Expected Fail: 1', lines)

    def test_history_show_failing(self):
        self.assertRunExit('stestr run passing', 0)
        self.assertRunExit('stestr run', 1)
        self.assertRunExit('stestr run passing', 0)
        output, _ = self.assertRunExit('stestr history show 1', 1)
        lines = [x.rstrip() for x in output.decode('utf8').split('\n')]
        self.assertIn(' - Passed: 2', lines)
        self.assertIn(' - Failed: 2', lines)
        self.assertIn(' - Expected Fail: 1', lines)
        self.assertIn(' - Unexpected Success: 1', lines)

    def test_history_show_invalid_id(self):
        self.assertRunExit('stestr run passing', 0)
        self.assertRunExit('stestr run', 1)
        self.assertRunExit('stestr run passing', 0)
        output, _ = self.assertRunExit('stestr history show 42', 1)
        self.assertEqual(output.decode('utf8').rstrip(), "'No such run.'")

    def test_history_remove(self):
        self.assertRunExit('stestr run passing', 0)
        self.assertRunExit('stestr run', 1)
        self.assertRunExit('stestr run passing', 0)
        self.assertRunExit('stestr history remove 1', 0)
        table = self.assertRunExit(
            'stestr history list', 0)[0].decode('utf8')
        self.assertIn("| 0      | True   |", table.split('\n')[3].rstrip())
        self.assertNotIn("| 1      | False  |", table.split('\n')[4].strip())
        self.assertIn("| 2      | True   |", table.split('\n')[4].rstrip())
        expected = """
+--------+--------+-----------+----------------------------------+
| Run ID | Passed | Runtime   | Date                             |
+--------+--------+-----------+----------------------------------+
""".strip()
        self.assertEqual(expected, '\n'.join(
            [x.rstrip() for x in table.split('\n')[:3]]))

    def test_no_subunit_trace_force_subunit_trace(self):
        out, err = self.assertRunExit(
            'stestr run --no-subunit-trace --force-subunit-trace passing', 0)
        out = str(out)
        self.assertNotIn('PASSED (id=0)', out)
        self.assertIn('Totals', out)
        self.assertIn('Worker Balance', out)
        self.assertIn('Sum of execute time for each test:', out)

    def test_parallel_passing_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(0, run.run_command(filters=['passing'],
                                            stdout=stdout.stream))

    def test_parallel_passing_bad_regex_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(1, run.run_command(filters=['bad.regex.foobar'],
                                            stdout=stdout.stream))

    def test_parallel_fails_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(1, run.run_command(stdout=stdout.stream))

    def test_serial_passing_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(0, run.run_command(filters=['passing'], serial=True,
                                            stdout=stdout.stream))

    def test_str_concurrency_passing_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(0, run.run_command(filters=['passing'],
                                            concurrency='1',
                                            stdout=stdout.stream))

    def test_str_concurrency_fails_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(1, run.run_command(concurrency='1',
                                            stdout=stdout.stream))

    def test_serial_fails_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(1, run.run_command(serial=True, stdout=stdout.stream))

    def test_serial_subunit_passing_from_func(self):
        stdout = io.BytesIO()
        self.assertEqual(0, run.run_command(subunit_out=True, serial=True,
                                            filters=['passing'],
                                            stdout=stdout))
        stdout.seek(0)
        self._check_subunit(stdout)

    def test_parallel_subunit_passing_from_func(self):
        stdout = io.BytesIO()
        self.assertEqual(0, run.run_command(subunit_out=True,
                                            filters=['passing'],
                                            stdout=stdout))
        stdout.seek(0)
        self._check_subunit(stdout)

    def test_until_failure_fails_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(1, run.run_command(until_failure=True,
                                            stdout=stdout.stream))

    def test_until_failure_with_subunit_fails_from_func(self):
        stdout = io.BytesIO()
        self.assertEqual(1, run.run_command(until_failure=True,
                                            subunit_out=True,
                                            stdout=stdout))
        stdout.seek(0)
        self._check_subunit(stdout)

    def test_list_from_func(self):
        stdout = fixtures.StringStream('stdout')
        self.useFixture(stdout)
        self.assertEqual(0, list_cmd.list_command(stdout=stdout.stream))

    def test_run_no_discover_pytest_path(self):
        passing_string = 'tests/test_passing.py::FakeTestClass::test_pass_list'
        out, err = self.assertRunExit('stestr run -n %s' % passing_string, 0)
        lines = out.decode('utf8').splitlines()
        self.assertIn(' - Passed: 1', lines)
        self.assertIn(' - Failed: 0', lines)

    def test_run_no_discover_pytest_path_failing(self):
        passing_string = 'tests/test_failing.py::FakeTestClass::test_pass_list'
        out, err = self.assertRunExit('stestr run -n %s' % passing_string, 1)
        lines = out.decode('utf8').splitlines()
        self.assertIn(' - Passed: 0', lines)
        self.assertIn(' - Failed: 1', lines)

    def test_run_no_discover_file_path(self):
        passing_string = 'tests/test_passing.py'
        out, err = self.assertRunExit('stestr run -n %s' % passing_string, 0)
        lines = out.decode('utf8').splitlines()
        self.assertIn(' - Passed: 2', lines)
        self.assertIn(' - Failed: 0', lines)
        self.assertIn(' - Expected Fail: 1', lines)

    def test_run_no_discover_file_path_failing(self):
        passing_string = 'tests/test_failing.py'
        out, err = self.assertRunExit('stestr run -n %s' % passing_string, 1)
        lines = out.decode('utf8').splitlines()
        self.assertIn(' - Passed: 0', lines)
        self.assertIn(' - Failed: 2', lines)
        self.assertIn(' - Unexpected Success: 1', lines)


class TestReturnCodesToxIni(TestReturnCodes):
    def setUp(self):
        super().setUp()
        self.tox_ini_dir = os.path.join(self.directory, 'tox.ini')
        tox_file = os.path.join(self.repo_root, 'stestr/tests/files/tox.ini')
        shutil.copy(tox_file, self.tox_ini_dir)
        os.remove(self.testr_conf_file)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import math

import testtools

from stestr import output


class IsolationAnalyzer:

    def __init__(self, latest_run, conf, run_func, repo, test_path=None,
                 top_dir=None, group_regex=None,
                 repo_url=None, serial=False, concurrency=0):
        super().__init__()
        self._worker_to_test = None
        self._test_to_worker = None
        self.latest_run = latest_run
        self.conf = conf
        self.group_regex = group_regex
        self.repo_url = repo_url
        self.serial = serial
        self.concurrency = concurrency
        self.test_path = test_path
        self.top_dir = top_dir
        self.run_func = run_func
        self.repo = repo

    def bisect_tests(self, spurious_failures):

        test_conflicts = {}
        if not spurious_failures:
            raise ValueError('No failures provided to bisect the cause of')
        for spurious_failure in spurious_failures:
            candidate_causes = self._prior_tests(self.latest_run,
                                                 spurious_failure)
            bottom = 0
            top = len(candidate_causes)
            width = top - bottom
            while width:
                check_width = int(math.ceil(width / 2.0))
                test_ids = candidate_causes[
                    bottom:bottom + check_width] + [spurious_failure]
                cmd = self.conf.get_run_command(
                    test_ids, group_regex=self.group_regex,
                    repo_url=self.repo_url,
                    serial=self.serial, concurrency=self.concurrency,
                    test_path=self.test_path, top_dir=self.top_dir)
                self.run_func(cmd, False,
                              pretty_out=False,
                              repo_url=self.repo_url)
                # check that the test we're probing still failed - still
                # awkward.
                found_fail = []

                def find_fail(test_dict):
                    if test_dict['id'] == spurious_failure:
                        found_fail.append(True)

                checker = testtools.StreamToDict(find_fail)
                checker.startTestRun()
                try:
                    self.repo.get_failing().get_test().run(checker)
                finally:
                    checker.stopTestRun()
                if found_fail:
                    # Our conflict is in bottom - clamp the range down.
                    top = bottom + check_width
                    if width == 1:
                        # found the cause
                        test_conflicts[
                            spurious_failure] = candidate_causes[bottom]
                        width = 0
                    else:
                        width = top - bottom
                else:
                    # Conflict in the range we did not run: discard bottom.
                    bottom = bottom + check_width
                    if width == 1:
                        # there will be no more to check, so we didn't
                        # reproduce the failure.
                        width = 0
                    else:
                        width = top - bottom
            if spurious_failure not in test_conflicts:
                # Could not determine cause
                test_conflicts[spurious_failure] = 'unknown - no conflicts'
        if test_conflicts:
            table = [('failing test', 'caused by test')]
            for failure in sorted(test_conflicts):
                causes = test_conflicts[failure]
                table.append((failure, causes))
            output.output_table(table)
            return 3
        return 0

    def _prior_tests(self, run, failing_id):
        """Calculate what tests from the test run run ran before test_id.

        Tests that ran in a different worker are not included in the result.
        """
        if not getattr(self, '_worker_to_test', False):
            case = run.get_test()
            # Use None if there is no worker-N tag
            # If there are multiple, map them all.
            # (worker-N -> [testid, ...])
            worker_to_test = {}
            # (testid -> [workerN, ...])
            test_to_worker = {}

            def map_test(test_dict):
                tags = test_dict['tags']
                id = test_dict['id']
                workers = []
                for tag in tags:
                    if tag.startswith('worker-'):
                        workers.append(tag)
                if not workers:
                    workers = [None]
                for worker in workers:
                    worker_to_test.setdefault(worker, []).append(id)
                test_to_worker.setdefault(id, []).extend(workers)

            mapper = testtools.StreamToDict(map_test)
            mapper.startTestRun()
            try:
                case.run(mapper)
            finally:
                mapper.stopTestRun()
            self._worker_to_test = worker_to_test
            self._test_to_worker = test_to_worker
        failing_workers = self._test_to_worker[failing_id]
        prior_tests = []
        for worker in failing_workers:
            worker_tests = self._worker_to_test[worker]
            prior_tests.extend(worker_tests[:worker_tests.index(failing_id)])
        return prior_tests



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import testtools


class TestCase(testtools.TestCase):

    def setUp(self):
        super().setUp()
        stdout = self.useFixture(fixtures.StringStream('stdout')).stream
        self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        stderr = self.useFixture(fixtures.StringStream('stderr')).stream
        self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))
        self.useFixture(fixtures.LoggerFixture(nuke_handlers=False,
                                               level=None))
